      PROGRAM M2HBO 
C********************************************************************* 
C***Usage
C     This program reads in a matrix A  stored in coordinate
C     format and writes out to a Harwell-Boeing formatted file.
C     The matrix is assumed to be stored as follows:
C     Each line contains the row number, the column number, and the 
C     real matrix entry. The row and column numbers can be written
C     in real form (like 1.00000e+00) and a line with a matrix entry
C     equal to zero will be used to determine the size of the matrix
C     but will further be ignored.
C     This program will read the files created by matlab using
C       > save filename A -ascii [-double]
C     where A is a real sparse matlab matrix
C
C***Author
C     John Haws, 6/30/99
C     Changed to accept direct matlab save files by :
C     Menno Verbeek, 8/13/99
C***Note: 
C     Matrix I/O from SPARSEKIT2 (Copyright (C) 1999 Yousef Saad)
C***Dependencies
C     readsm, prtmt
C***Usage
C     Let matrix.mtx be a file containing matrix entries stored in
C     coordinate format (could be direct save file from matlab, 
C     see above) The command
C        m2hbo < matrix.mtx > matrix.rua
C     writes the matrix out to the file matrix.rua in Harwell-Boeing
C     format.
C
C**********************************************************************
C
C***Parameters
C   ---------
C **Integer
C     N        = order of the matrix
C     NELT     = number of non-zeros stored in matrix
C     JOB      = integer indicating whether matrix values and a rhs
C                are to be written out. See prtmt routine below.
C     IFMT     = integer specifying format for real value output
C                see prtmt routine below
C **Integer Array
C     JA       = integer array of the same length as a containing 
C                the column indices of the corresponding matrix 
C                elements of array a.
C     IA       = integer array of containing the pointers to the 
C                beginning of the row in arrays a and ja.
C **Double Precision
C     A(*)      = stores non-zero entries of matrix
C **I/O
C     IIN, IOUT = input and output unit numbers (set to 5=std.in,
C                 6=std.out, respectively)
C
C*****************************************************************
C
      INTEGER           NMAX, NELTMAX 
      PARAMETER ( NMAX = 1000000, NELTMAX = 5000000 )
      INTEGER           N, NELT, JOB, IFMT
      INTEGER           JA(NELTMAX), IA(NMAX+1)
      DOUBLE PRECISION  A(NELTMAX)

C
C -- I/O Arguments
C
      DATA              IIN /5/, IOUT/6/
      CHARACTER         TITLE*72 , KEY*8 , TYPE*3
C
C -- Externals
C
      EXTERNAL          READSM, PRTMT
C
C-----------------------------------------------------------------
C -- Begin Main Driver
C-----------------------------------------------------------------
C
C -- Read input matrix
C
      CALL READSM ( NMAX, NELTMAX, N, NELT, IA, JA, A, IIN, IERR)
      IF (IERR.NE.0) WRITE(0,*) 'Problem reading file, ierr=',ierr
C
C -- Initialize some dummy arguments
C
      TITLE='OUTFILE FROM COORDINATE/MATLAB FILE'
      KEY='OUTFILE'
C ---- R=>real, U=>nonsymmetric, A=>assembled
      TYPE='RUA'
#ifdef USE_DOUBLE
      IFMT=17
#else
      IFMT=8
#endif
C ---- JOB=2 => write out matrix, no rhs
      JOB=2
C
C -- Write output matrix
C
      CALL PRTMT( N, N, A, JA, IA, RHS, GUESOL, TITLE, 
     &       KEY, TYPE, IFMT, JOB, IOUT )

      STOP
C
C -- End of main driver program
C
      END
C=======================================================================
C Begin Saad's routines for I/O, conversion
C=======================================================================
c-----------------------------------------------------------------------
      subroutine readsm (nmax,nzmax,n,nnz,ia,ja,a,iout,ierr)
      integer nmax, nzmax, row, n, iout, i, j, k, ierr
      integer ia(nmax+1), ja(nzmax)
      double precision  a(nzmax), x
      double precision  di,dj
c-----------------------------------------------------------------------
c     read a matrix in coordinate format as is used in the SMMS
c     package (F. Alvarado), i.e. the row is in ascending order.
c     Outputs the matrix in CSR format.
c-----------------------------------------------------------------------
c coded by Kesheng Wu on Oct 21, 1991 with the supervision of Y. Saad
c-----------------------------------------------------------------------
c on entry:
c---------
c nmax  = the maximum size of array
c nzmax = the maximum number of nonzeros
c iout  = the I/O unit that has the data file
c
c on return:
c----------
c n     = integer = size of matrix
c nnz   = number of non-zero entries in the matrix
c a,
c ja,
c ia    = matrix in CSR format
c ierr  = error code,
c         0 -- subroutine end with intended job done
c         1 -- error in I/O unit iout
c         2 -- end-of-file reached while reading n, 
c              i.e. a empty data file
c         3 -- n non-positive or too large
c         4 -- nnz is zero or larger than nzmax
c         5 -- data file is not orgnized in the order of ascending
c              row indices
c
c in case of errors:
c   n will be set to zero (0). In case the data file has more than nzmax
c   number of entries, the first nzmax entries will be read, and are not
c   cleared on return. The total number of entry is determined.
c   Ierr is set.
c-----------------------------------------------------------------------
c
      rewind(iout)
      nnz = 0
      ia(1) = 1
      row = 1
      n = 0
c
c      read (iout,*, err=1000, end=1010) n
c      if ((n.le.0) .or. (n.gt.nmax)) goto 1020
c

 10   continue

c      read (iout, *, err=1000, end=100) j, i, x
      read (iout, *, err=1000, end=100) dj, di, x
      i=int(di)
      j=int(dj)
      n=max(n,i)
      n=max(n,j)
      if (n.gt.nmax) goto 1020

      nnz = nnz + 1

c     set the pointers when needed
      if (i.gt.row) then
         do 20 k = row+1, i
            ia(k) = nnz
 20      continue
         row = i
      else if (i.lt.row) then
         goto 1040
      endif
 
      if (x .eq. 0.0d0) then
         nnz = nnz - 1
         goto 10
      end if

      ja(nnz) = j
      a (nnz) = x
 
      if (nnz.lt.nzmax) then
         goto 10
      else
         goto 1030
      endif
 
c     normal return -- end of file reached
 100  continue
      ia(row+1) = nnz + 1
c      if (nnz.eq.0) goto 1030
c
c     everything seems to be OK.
c
      ierr = 0
      return
c
c     error handling code
c
c     error in reading data entries
c
 1000 ierr = 1
      WRITE(0,*) '*** Error reading data entries ***'
      goto 2000
c
c     empty file
c
 1010 ierr  = 2
      WRITE(0,*) '*** Empty file ***'
      goto 2000
c
c     problem with n
c
 1020 ierr = 3
      WRITE(0,*) '*** Problem with N, order of matrix ***'
      goto 2000
c
c     problem with nnz
c
 1030 ierr = 4
      WRITE(0,*) '*** Problem with NNZ, number of entries ***'
c
c     try to determine the real number of entries, in case needed
c
      if (nnz.ge.nzmax) then
 200     read(iout, *, err=210, end=210) i, j, x
         nnz = nnz + 1
         goto 200
 210     continue
      endif
      goto 2000
c
c     data entries not ordered
c
 1040 ierr = 5
      WRITE(0,*) '*** Data entries not ordered ***'
 2000 n = 0
      return
c----end-of-readsm------------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine prtmt (nrow,ncol,a,ja,ia,rhs,guesol,title,key,type,
     1     ifmt,job,iounit)
c-----------------------------------------------------------------------
c writes a matrix in Harwell-Boeing format into a file.
c assumes that the matrix is stored in COMPRESSED SPARSE COLUMN FORMAT.
c some limited functionality for right hand sides.
c Author: Youcef Saad - Date: Sept., 1989 - updated Oct. 31, 1989 to
c cope with new format.
c-----------------------------------------------------------------------
c on entry:
c---------
c nrow   = number of rows in matrix
c ncol   = number of columns in matrix
c a      = real*8 array containing the values of the matrix stored
c          columnwise
c ja     = integer array of the same length as a containing the column
c          indices of the corresponding matrix elements of array a.
c ia     = integer array of containing the pointers to the beginning of
c          the row in arrays a and ja.
c rhs    = real array  containing the right-hand-side (s) and optionally
c          the associated initial guesses and/or exact solutions
c          in this order. See also guesol for details. the vector rhs
c          will be used only if job .gt. 2 (see below). Only full 
c          storage for the right hand sides is supported.
c
c guesol = a 2-character string indicating whether an initial guess
c          (1-st character) and / or the exact solution (2-nd)
c          character) is provided with the right hand side.
c          if the first character of guesol is 'G' it means that an
c          an intial guess is provided for each right-hand sides.
c          These are assumed to be appended to the right hand-sides in
c          the array rhs.
c          if the second character of guesol is 'X' it means that an
c          exact solution is provided for each right-hand side.
c          These are assumed to be appended to the right hand-sides
c          and the initial guesses (if any) in the array rhs.
c
c title  = character*72 = title of matrix test ( character a*72 ).
c key    = character*8  = key of matrix
c type   = charatcer*3  = type of matrix.
c
c ifmt   = integer specifying the format chosen for the real values
c          to be output (i.e., for a, and for rhs-guess-sol if
c          applicable). The meaning of ifmt is as follows.
c         * if (ifmt .lt. 100) then the D descriptor is used,
c           format Dd.m, in which the length (m) of the mantissa is
c           precisely the integer ifmt (and d = ifmt+6)
c         * if (ifmt .gt. 100) then prtmt will use the
c           F- descriptor (format Fd.m) in which the length of the
c           mantissa (m) is the integer mod(ifmt,100) and the length
c           of the integer part is k=ifmt/100 (and d = k+m+2)
c           Thus  ifmt= 4   means  D10.4  +.xxxxD+ee    while
c                 ifmt=104  means  F7.4   +x.xxxx
c                 ifmt=205  means  F9.5   +xx.xxxxx
c           Note: formats for ja, and ia are internally computed.
c
c job    = integer to indicate whether matrix values and
c          a right-hand-side is available to be written
c          job = 1   write srtucture only, i.e., the arrays ja and ia.
c          job = 2   write matrix including values, i.e., a, ja, ia
c          job = 3   write matrix and one right hand side: a,ja,ia,rhs.
c          job = nrhs+2 write matrix and nrhs successive right hand 
c          sides. Note that there cannot be any right-hand-side if the 
c          matrix has no values. Also the initial guess and exact 
c          solutions when provided are for each right hand side. For 
c          example if nrhs=2 and guesol='GX' there are 6 vectors to 
c          write.
c
c
c iounit = logical unit number where to write the matrix into.
c
c on return:
c----------
c the matrix a, ja, ia will be written in output unit iounit
c in the Harwell-Boeing format. None of the inputs is modofied.
c
c Notes: 1) This code attempts to pack as many elements as possible per
c        80-character line.
c        2) this code attempts to avoid as much as possible to put
c        blanks in the formats that are written in the 4-line header
c        (This is done for purely esthetical reasons since blanks
c        are ignored in format descriptors.)
c        3) sparse formats for right hand sides and guesses are not
c        supported.
c-----------------------------------------------------------------------
      character title*72,key*8,type*3,ptrfmt*16,indfmt*16,valfmt*20,
     *          guesol*2, rhstyp*3
      integer totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol,
     1     nnz, nrhs, len, nperli, nrwindx
      integer ja(*), ia(*)
      double precision a(*),rhs(*)
c--------------
c     compute pointer format
c--------------
      nnz    = ia(ncol+1) -1
      len    = int ( alog10(0.1+real(nnz+1))) + 1
      nperli = 80/len
      ptrcrd = ncol/nperli + 1
      if (len .gt. 9) then
         assign 101 to ix
      else
         assign 100 to ix
      endif
      write (ptrfmt,ix) nperli,len
 100  format(1h(,i2,1HI,i1,1h) )
 101  format(1h(,i2,1HI,i2,1h) )
c----------------------------
c compute ROW index format
c----------------------------
      len    = int ( alog10(0.1+real(nrow) )) + 1
      nperli = min0(80/len,nnz)
      indcrd = (nnz-1)/max(1,nperli)+1
      write (indfmt,100) nperli,len
c---------------
c compute values and rhs format (using the same for both)
c---------------
      valcrd    = 0
      rhscrd  = 0
c quit this part if no values provided.
      if (job .le. 1) goto 20
c
      if (ifmt .ge. 100) then
         ihead = ifmt/100
         ifmt = ifmt-100*ihead
         len = ihead+ifmt+2
      print*, len
         nperli = 80/len
c
         if (len .le. 9 ) then
            assign 102 to ix
         elseif (ifmt .le. 9) then
            assign 103 to ix
         else
            assign 104 to ix
         endif
c
         write(valfmt,ix) nperli,len,ifmt
 102     format(1h(,i2,1hF,i1,1h.,i1,1h) )
 103     format(1h(,i2,1hF,i2,1h.,i1,1h) )
 104     format(1h(,i2,1hF,i2,1h.,i2,1h) )
C
      else
         len = ifmt + 6
         nperli = 80/len
c     try to minimize the blanks in the format strings.
         if (nperli .le. 9) then
            if (len .le. 9 ) then
               assign 105 to ix
            elseif (ifmt .le. 9) then
               assign 106 to ix
            else
               assign 107 to ix
            endif
         else
            if (len .le. 9 ) then
               assign 108 to ix
            elseif (ifmt .le. 9) then
               assign 109 to ix
            else
               assign 110 to ix
            endif
         endif
c-----------
         write(valfmt,ix) nperli,len,ifmt
 105     format(1h(,i1,1hD,i1,1h.,i1,1h) )
 106     format(1h(,i1,1hD,i2,1h.,i1,1h) )
 107     format(1h(,i1,1hD,i2,1h.,i2,1h) )
 108     format(1h(,i2,1hD,i1,1h.,i1,1h) )
 109     format(1h(,i2,1hD,i2,1h.,i1,1h) )
 110     format(1h(,i2,1hD,i2,1h.,i2,1h) )
c
      endif
      valcrd = (nnz-1)/nperli+1
      nrhs   = job -2
      if (nrhs .ge. 1) then
         i = (nrhs*nrow-1)/nperli+1
         rhscrd = i
         if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g')
     +      rhscrd = rhscrd+i
         if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x')
     +      rhscrd = rhscrd+i
         rhstyp = 'F'//guesol
      endif
 20   continue
c
      totcrd = ptrcrd+indcrd+valcrd+rhscrd
c     write 4-line or five line header
      write(iounit,10) title,key,totcrd,ptrcrd,indcrd,valcrd,
     1     rhscrd,type,nrow,ncol,nnz,nrhs,ptrfmt,indfmt,valfmt,valfmt
c-----------------------------------------------------------------------
      nrwindx = 0
      if (nrhs .ge. 1) write (iounit,11) rhstyp, nrhs, nrwindx
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)
 11   format(A3,11x,i14,i14)
c
      write(iounit,ptrfmt) (ia (i), i = 1, ncol+1)
      write(iounit,indfmt) (ja (i), i = 1, nnz)
      if (job .le. 1) return
      write(iounit,valfmt) (a(i), i = 1, nnz)
      if (job .le. 2) return
      len = nrow*nrhs
      next = 1
      iend = len
      write(iounit,valfmt) (rhs(i), i = next, iend)
c
c     write initial guesses if available
c
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
         next = next+len
         iend = iend+ len
         write(iounit,valfmt) (rhs(i), i = next, iend)
      endif
c
c     write exact solutions if available
c
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') then
         next = next+len
         iend = iend+ len
         write(iounit,valfmt) (rhs(i), i = next, iend)
      endif
c
      return
c----------end of prtmt ------------------------------------------------
c-----------------------------------------------------------------------
      end
